# -*- coding: utf-8 -*-
# ===============LICENSE_START=======================================================
# Acumos Apache-2.0
# ===================================================================================
# Copyright (C) 2017-2018 AT&T Intellectual Property & Tech Mahindra. All rights reserved.
# ===================================================================================
# This Acumos software file is distributed by AT&T and Tech Mahindra
# under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============LICENSE_END=========================================================
"""
Provides custom pickle utilities
"""
import sys
import json
import inspect
import contextlib
import tempfile
from os import makedirs
from os.path import basename, isdir, isfile, join as path_join
from copy import deepcopy
from functools import partial
from typing import GenericMeta, Dict, List
from types import ModuleType
from importlib import import_module

import dill

from .modeling import _is_namedtuple, create_namedtuple, Empty
from .utils import namedtuple_field_types


_DEFAULT_MODULES = ('ucumos', 'dill')
_BLACKLIST = {'builtins', }
_DEFAULT = 'default'

_contexts = dict()

dump_model = partial(dill.dump, recurse=True)
dumps_model = partial(dill.dumps, recurse=True)
load_model = dill.load
loads_model = dill.loads


def _save_annotation(pickler, obj):
    '''Workaround for dill annotation serialization bug'''
    if obj.__origin__ in (Dict, List):
        # recursively save object
        t = obj.__origin__
        args = obj.__args__
        pickler.save_reduce(_load_annotation, (t, args), obj=obj)
    else:
        # eventually hit base type, then use stock pickling logic. temp revert prevents infinite recursion
        t = obj
        args = None
        with _revert_dispatch(GenericMeta):
            pickler.save_reduce(_load_annotation, (t, args), obj=obj)


def _load_annotation(t, args):
    '''Workaround for dill annotation serialization bug'''
    if t is Dict and args is not None:
        return Dict[args[0], args[1]]
    elif t is List and args is not None:
        return List[args[0]]
    else:
        return t


def _save_namedtuple(pickler, obj):
    '''Workaround for dill NamedTuple serialization bug'''
    field_types = namedtuple_field_types(obj)
    pickler.save_reduce(_load_namedtuple, (obj.__name__, field_types), obj=obj)


def _load_namedtuple(name, field_types):
    '''Workaround for dill NamedTuple serialization bug'''
    return create_namedtuple(name, [(k, v) for k, v in field_types.items()])


@contextlib.contextmanager
def _revert_dispatch(t):
    '''Temporarily removes a type from the dispatch table to use existing serialization logic'''
    f = dill.Pickler.dispatch.pop(t)
    try:
        yield
    finally:
        dill.Pickler.dispatch[t] = f


def _add_file(subdir, name):
    '''Helper function which returns the absolute and context-relative path of a file to be added'''
    file_abspath = path_join(subdir, name)
    file_relpath = (basename(subdir), name)
    return file_abspath, file_relpath


dill.Pickler.dispatch[GenericMeta] = _save_annotation


_CUSTOM_DISPATCH = {
}


@contextlib.contextmanager
def _patch_dill():
    '''Temporarily patches the dill Pickler dispatch table to support custom serialization within a context'''
    try:
        dispatch = dill.Pickler.dispatch
        dill.Pickler.dispatch = deepcopy(dispatch)

        pickler_save = dill.Pickler.save

        def wrapped_save(pickler, obj, save_persistent_id=True):
            '''Hook that intercepts objects about to be saved'''
            _catch_object(obj)

            if _is_namedtuple(obj) and obj is not Empty:
                _save_namedtuple(pickler, obj)
            else:
                pickler_save(pickler, obj, save_persistent_id)

        dill.Pickler.save = wrapped_save
        yield
    finally:
        dill.Pickler.dispatch = dispatch
        dill.Pickler.save = pickler_save


def _catch_object(obj):
    '''Inspects object and executes custom serialization / bookkeeping logic'''

    # dynamically extend dispatch table to prevent unnecessary imports / dependencies
    obj_type = obj if inspect.isclass(obj) else type(obj)
    if obj_type not in dill.Pickler.dispatch:
        for path in _get_mro_paths(obj_type):
            if path in _CUSTOM_DISPATCH:
                dill.Pickler.dispatch[obj_type] = _CUSTOM_DISPATCH[path]

    base_module = _get_base_module(obj)
    if base_module is not None and base_module.__name__ not in _BLACKLIST:
        context = get_context()
        context.add_module(base_module)


def _get_base_module(obj):
    '''Returns the base module for a given object'''
    module = inspect.getmodule(obj)
    if module is not None:
        base_name, _, _ = module.__name__.partition('.')
        base_module = sys.modules[base_name]
    else:
        base_module = None
    return base_module


def _get_mro_paths(type_):
    '''Yields import path string for each entry in `inspect.getmro`'''
    for t in inspect.getmro(type_):
        yield "{}.{}".format(t.__module__, t.__name__)


class AcumosContext(object):
    '''Represents a workspace for a model that is being dumped'''

    def __init__(self, root_dir):
        if not isdir(root_dir):
            raise Exception("AcumosContext root directory {} does not exist".format(root_dir))
        self._modules = set()
        self._root_dir = root_dir
        self._params_path = path_join(root_dir, 'context.json')
        self.parameters = self._load_params()

        for mod in _DEFAULT_MODULES:
            self.add_module(mod)

    def create_subdir(self, *paths, exist_ok=False):
        '''Creates a new directory within the context root and returns the absolute path'''
        if not paths:
            tdir = tempfile.mkdtemp(dir=self._root_dir)
        else:
            tdir = path_join(self._root_dir, *paths)
            makedirs(tdir, exist_ok=exist_ok)
        return tdir

    def build_path(self, *paths):
        '''Returns an absolute path starting from the context root'''
        return path_join(self._root_dir, *paths)

    def add_module(self, module):
        '''Adds a module to the context module set'''
        if isinstance(module, str):
            try:
                module = import_module(module)
            except ImportError:
                raise Exception("Module '{}' was identified as a dependency, but cannot be imported. Ensure that it is installed and available".format(module))
        elif not isinstance(module, ModuleType):
            raise Exception("Module must be of type str or types.ModuleType, not {}".format(type(module)))

        self._modules.add(module)

    @property
    def abspath(self):
        '''Absolute path of the context root directory'''
        return self._root_dir

    @property
    def basename(self):
        '''Base name of the context root directory'''
        return basename(self._root_dir)

    @property
    def modules(self):
        '''The set of all modules (i.e. typing.ModuleType) identified as dependencies'''
        return frozenset(self._modules)

    @property
    def packages(self):
        '''The set of all base packages (i.e. typing.ModuleType with a package) identified as dependencies'''
        return frozenset(mod for mod in self._modules if mod.__package__)

    @property
    def package_names(self):
        '''The set of all base package names identified as dependencies'''
        return frozenset(mod.__name__ for mod in self.packages)

    @property
    def scripts(self):
        '''The set of all scripts (i.e. typing.ModuleType with no package) identified as dependencies'''
        return frozenset(script for script in (mod for mod in self._modules if not mod.__package__) if script.__name__ != '__main__')

    @property
    def script_names(self):
        '''The set of all script names identified as dependencies'''
        return frozenset(mod.__name__ for mod in self.scripts)

    def _load_params(self):
        '''Returns a parameters dict'''
        if isfile(self._params_path):
            with open(self._params_path) as f:
                return json.load(f)
        else:
            return dict()

    def save_params(self):
        '''Saves a parameters json file within the context root'''
        with open(self._params_path, 'w') as f:
            json.dump(self.parameters, f)


@contextlib.contextmanager
def AcumosContextManager(rootdir=None, name=_DEFAULT):
    '''Context manager that provides a AcumosContext object'''
    with _patch_dill():
        if name in _contexts:
            raise Exception("AcumosContext '{}' has already been created. Use `get_context` to access it.".format(name))
        try:
            with _DirManager(rootdir) as rootdir:
                context = AcumosContext(rootdir)
                _contexts[name] = context
                yield context
                context.save_params()
        finally:
            del _contexts[name]


@contextlib.contextmanager
def _DirManager(dir_=None):
    '''Wrapper that passes dir_ through or creates a temporary directory'''
    if dir_ is not None:
        if not isdir(dir_):
            raise Exception("Provided AcumosContext rootdir {} does not exist".format(dir_))
        yield dir_
    else:
        with tempfile.TemporaryDirectory() as tdir:
            yield tdir


def get_context(name=_DEFAULT):
    '''Returns an existing AcumosContext'''
    if name in _contexts:
        return _contexts[name]
    else:
        raise Exception("AcumosContext '{}' has not been created".format(name))



# -*- coding: utf-8 -*-
# ===============LICENSE_START=======================================================
# Acumos Apache-2.0
# ===================================================================================
# Copyright (C) 2017-2018 AT&T Intellectual Property & Tech Mahindra. All rights reserved.
# ===================================================================================
# This Acumos software file is distributed by AT&T and Tech Mahindra
# under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============LICENSE_END=========================================================
"""
Provides wrapped model utilities
"""
import os
import sys
from zipfile import ZipFile
from os import listdir
from os.path import isfile, isdir, join as path_join

from google.protobuf.json_format import Parse as ParseJson, ParseDict, MessageToJson, MessageToDict

from .modeling import _is_namedtuple, List, Dict
from .pickler import AcumosContextManager, load_model as _load_model
from .utils import load_module


def load_packaged_model(path):
    '''Returns a Model previously dumped to `path`'''
    model_zip_path = path_join(path, 'model.zip')
    model_dir = path_join(path, 'model')
    ZipFile(model_zip_path).extractall(model_dir)

    with AcumosContextManager(model_dir) as c:
        _extend_syspath(c)  # extend path before we unpickle user model
        with open(c.build_path('model.pkl'), 'rb') as f:
            model = _load_model(f)
    os.system('rm -rf ' + model_dir)
    return model


def load_model(path):
    '''Returns a WrappedModel previously dumped to `path`'''
    model_dir = _infer_model_dir(path)
    with AcumosContextManager(model_dir) as c:
        _extend_syspath(c)  # extend path before we unpickle user model

        with open(c.build_path('model.pkl'), 'rb') as f:
            model = _load_model(f)

        pkg = c.parameters['protobuf_package']
        module_path = c.build_path(path_join('scripts', 'acumos_gen', pkg, 'model_pb2.py'))
        module = load_module('model_pb2', module_path)
        return WrappedModel(model, module)


def _infer_model_dir(path):
    '''Returns an absolute path to the model dir. Unzips the model archive if `path` contains it'''
    model_zip_path = path_join(path, 'model.zip')
    if isfile(model_zip_path):
        model_dir = path_join(path, 'model')
        zip_file = ZipFile(model_zip_path)
        zip_file.extractall(model_dir)
    else:
        model_dir = path

    pkl_path = path_join(model_dir, 'model.pkl')
    if not isfile(pkl_path):
        raise Exception("Provided path {} does not contain an Acumos model".format(path))

    return model_dir


def _extend_syspath(context):
    '''Adds user-provided packages to the system path'''
    provided_abspath = context.build_path(path_join('scripts', 'user_provided'))
    if provided_abspath not in sys.path:
        sys.path.append(provided_abspath)

    for pkg_name in listdir(provided_abspath):
        pkg_abspath = path_join(provided_abspath, pkg_name)
        if not isdir(pkg_abspath):
            continue

        if pkg_abspath not in sys.path:
            sys.path.append(pkg_abspath)


class WrappedModel(object):
    '''Container of WrappedFunction objects'''

    def __init__(self, model, module):
        self._methods = {name: WrappedFunction(func, module) for name, func in model.methods.items()}
        for k, v in self._methods.items():
            setattr(self, k, v)

    @property
    def methods(self):
        return self._methods


class WrappedFunction(object):
    '''A function wrapper with various consumption options'''

    def __init__(self, func, module):
        self._func = func
        self._module = module
        self._input_type = func.input_type
        self._output_type = func.output_type
        self._pb_input_type = getattr(module, func.input_type.__name__)
        self._pb_output_type = getattr(module, func.output_type.__name__)

    def from_pb_bytes(self, pb_bytes_in):
        '''Consumes a binary Protobuf message and returns a WrappedResponse object'''
        pb_msg_in = self._pb_input_type.FromString(pb_bytes_in)
        return self.from_pb_msg(pb_msg_in)

    def from_pb_msg(self, pb_msg_in):
        '''Consumes a Protobuf message object and returns a WrappedResponse object'''
        wrapped_in = _unpack_pb_msg(self._input_type, pb_msg_in)
        return self.from_wrapped(wrapped_in)

    def from_native(self, *args, **kwargs):
        '''Consumes the original inner function arguments and returns a WrappedResponse object'''
        wrapped_in = self._input_type(*args, **kwargs)
        return self.from_wrapped(wrapped_in)

    def from_wrapped(self, wrapped_in):
        '''Consumes a NamedTuple wrapped type and returns a WrappedResponse object'''
        wrapped_out = self._func.wrapped(wrapped_in)
        return WrappedResponse(wrapped_out, self._module, self._pb_output_type)

    def from_dict(self, dict_in):
        '''Consumes a dict and returns a WrappedResponse object'''
        pb_msg_in = ParseDict(dict_in, self._pb_input_type())
        return self.from_pb_msg(pb_msg_in)

    def from_json(self, json_in):
        '''Consumes a json str and returns a WrappedResponse object'''
        pb_msg_in = ParseJson(json_in, self._pb_input_type())
        return self.from_pb_msg(pb_msg_in)

    @property
    def pb_input_type(self):
        return self._pb_input_type

    @property
    def pb_output_type(self):
        return self._pb_output_type


class WrappedResponse(object):
    '''A WrappedFunction response with various return options'''

    def __init__(self, resp, module, pb_output_type):
        self._resp = resp
        self._module = module
        self._pb_output_type = pb_output_type

    def as_pb_bytes(self):
        '''Returns a Protobuf binary string representation of the model response'''
        return self.as_pb_msg().SerializeToString()

    def as_pb_msg(self):
        '''Returns a Protobuf message representation of the model response'''
        return _pack_pb_msg(self._resp, self._module)

    def as_wrapped(self):
        '''Returns a Python NamedTuple representation of the model response'''
        return self._resp

    def as_dict(self):
        '''Returns a dict representation of the model response'''
        pb_msg_out = self.as_pb_msg()
        return MessageToDict(pb_msg_out, self._pb_output_type())

    def as_json(self):
        '''Returns a json str representation of the model response'''
        pb_msg_out = self.as_pb_msg()
        return MessageToJson(pb_msg_out, self._pb_output_type(), indent=0)


def _pack_pb_msg(wrapped_in, module):
    '''Returns a protobuf message object from a NamedTuple instance'''
    wrapped_type = type(wrapped_in)
    field_types = wrapped_type._field_types
    pb_type = getattr(module, wrapped_type.__name__)
    return pb_type(**{f: _set_pb_value(field_types[f], v, module) for f, v in zip(wrapped_in._fields, wrapped_in)})


def _set_pb_value(wrapped_type, value, module):
    '''Recursively traverses the NamedTuple instance to ensure nested NamedTuples become protobuf messages'''
    if _is_namedtuple(wrapped_type):
        return _pack_pb_msg(value, module)

    elif issubclass(wrapped_type, Dict):
        _, val_type = wrapped_type.__args__
        if _is_namedtuple(val_type):
            return {k: _pack_pb_msg(v, module) for k, v in value.items()}

    elif issubclass(wrapped_type, List):
        list_type = wrapped_type.__args__[0]
        if _is_namedtuple(list_type):
            return [_pack_pb_msg(v, module) for v in value]

    return value


def _unpack_pb_msg(input_type, pb_msg):
    '''Returns a NamedTuple from protobuf message'''
    values = {f: _get_pb_value(t, getattr(pb_msg, f)) for f, t in input_type._field_types.items()}
    return input_type(**values)


def _get_pb_value(wrapped_type, pb_value):
    '''Recursively traverses the protobuf message to ensure nested messages become NamedTuples'''
    if _is_namedtuple(wrapped_type):
        return _unpack_pb_msg(wrapped_type, pb_value)

    elif issubclass(wrapped_type, Dict):
        _, val_type = wrapped_type.__args__
        if _is_namedtuple(val_type):
            return {k: _unpack_pb_msg(val_type, v) for k, v in pb_value.items()}

    elif issubclass(wrapped_type, List):
        list_type = wrapped_type.__args__[0]
        if _is_namedtuple(list_type):
            return [_unpack_pb_msg(list_type, v) for v in pb_value]

    return pb_value
